"""Module with spark native feature importance score tools.
"""
from typing import Union
from typeguard import typechecked
from numpy import exp as np_exp
import pandas as pd

import pyspark


@typechecked
def extract_features_score(
    model: Union[
        pyspark.ml.classification.LogisticRegressionModel,
        pyspark.ml.classification.DecisionTreeClassificationModel,
        pyspark.ml.classification.RandomForestClassificationModel,
        pyspark.ml.classification.GBTClassificationModel,
        pyspark.ml.regression.LinearRegressionModel,
        pyspark.ml.regression.DecisionTreeRegressionModel,
        pyspark.ml.regression.RandomForestRegressionModel,
        pyspark.ml.regression.GBTRegressionModel
    ],
    dfs: pyspark.sql.dataframe.DataFrame,
    features_col: str = 'features'
) -> pd.core.frame.DataFrame:
    """Function that extracts feature importance or coefficients from spark models.

    There are 3 possible situations created inside this function:
        1) If model is a DecisionTree, RandomForest or GBT (classification or regression) the features score will be gini scores.
        See https://spark.apache.org/docs/latest/api/python/reference/api/pyspark.ml.classification.DecisionTreeClassificationModel.html?highlight=featureimportances#pyspark.ml.classification.DecisionTreeClassificationModel.featureImportances
        for a description on the feature importance;
        
        2) If model is a LogisticRegression the features scores will be the odds ratio;

        3) If model is a LinearRegression the features socres will be the coefficients.

    Args:
        model (Union[spark models]): A fitted spark model. Accepted models: DecisionTree, RandomForest, GBT (both classification and regression), LogisticRegression and LinearRegression.
        dfs (pyspark.sql.dataframe.DataFrame): The SparkDataFrame used to fit the model.
        features_col (str, optional): The features vector column. Defaults to 'features'.

    Raises:
        ValueError: if features_col not in dfs.columns is True.

    Returns:
        [pd.core.frame.DataFrame]: A PandasDataFrame with the folloing columns:
            - feat_index: The feature index number in the features vector column;
            - feature: The feature name;
            - delta_gini/odds_ratio/coefficient: The feature score depending on the model (see the description).
    """
    if features_col not in dfs.columns:
        raise ValueError(f'features_col "{features_col}" not in dfs.columns.')

    list_extract = []
    # If regression then get coefficients else get featureImportances (gini)
    if type(model) in (pyspark.ml.classification.LogisticRegressionModel, pyspark.ml.regression.LinearRegressionModel):
        featureImp = model.coefficients
    else:
        featureImp = model.featureImportances

    for i in dfs.schema[features_col].metadata["ml_attr"]["attrs"]:
        list_extract = list_extract + dfs.schema[features_col].metadata["ml_attr"]["attrs"][i]
    
    df_fi = pd.DataFrame(list_extract)
    df_fi['score'] = df_fi['idx'].apply(lambda x: featureImp[x])
    df_fi = df_fi.sort_values('score', ascending = False)
    if 'vals' in list(df_fi.columns):
        df_fi.drop(columns=['vals'], inplace=True)

    if type(model) == pyspark.ml.classification.LogisticRegressionModel:
        df_fi.columns = ['feat_index', 'feature', 'odds_ratio']
        df_fi['odds_ratio'] = np_exp(df_fi['odds_ratio'])
    
    elif type(model) == pyspark.ml.regression.LinearRegressionModel:
        df_fi.columns = ['feat_index', 'feature', 'coefficients']
    
    else:
        df_fi.columns = ['feat_index', 'feature', 'delta_gini']
    
    return df_fi



from typing import Tuple, List, Union
from typeguard import typechecked
import pandas as pd

from pyspark import keyword_only
from pyspark.sql import DataFrame, functions as F, types as T, SparkSession
from pyspark.ml import Transformer
from pyspark.ml.feature import QuantileDiscretizer
from pyspark.ml.param import Param, Params, TypeConverters
from pyspark.ml.param.shared import HasInputCol, HasInputCols, HasOutputCol, HasOutputCols
from pyspark.ml.util import DefaultParamsReadable, DefaultParamsWritable

from pyspark_ds_toolbox.ml.data_prep.features_vector import get_features_vector

@typechecked
def compute_woe_iv(
    dfs: DataFrame,
    col_feature: str,
    col_target: str
) -> Tuple[DataFrame, float]:
    """Function that given a DataFrame, a categorical feature and a binary target column computes the Information Value.
    
    See http://www.m-hikari.com/ams/ams-2014/ams-65-68-2014/zengAMS65-68-2014.pdf for a technical description.

    Args:
        dfs (DataFrame): A spark DataFrame with col_feature and col_target;
        col_feature (str): Column name of a categorical feature.
        col_target (str): Column name of a binary target column. Must be of integer type and have only value 0 and 1.

    Raises:
        TypeError: if dfs.schema[col_target].dataType != T.IntegerType
        ValueError: unique_target_values != [0, 1]

    Returns:
        Tuple[DataFrame, float]: A two element tuple with the follwing objects:
            - Spark DataFrame with 'feature', 'feature_value', 'woe', 'iv' column;
            - float with the col_feature information value.
    """
    if dfs.schema[col_target].dataType != T.IntegerType():
        raise TypeError(f'Column {col_target} is of type {dfs.schema[col_target].dataType}. Must be IntegerType.')
    
   
    unique_target_values = dfs.select(col_target).distinct().collect()
    unique_target_values = [v[col_target] for v in unique_target_values]
    unique_target_values.sort()
    allTrue = all([(x==0) or (x==1) for x in unique_target_values])    
    if allTrue == False:
        raise ValueError(f'Unique values from {col_target} column are not [0, 1]. ({unique_target_values})')
    
     
    cross = dfs\
        .crosstab(col_feature, col_target)\
        .withColumnRenamed(f'{col_feature}_{col_target}', 'feature_value')
        
    sum_0 = cross.select('0').groupBy().sum().collect()[0][0]
    sum_1 = cross.select('1').groupBy().sum().collect()[0][0]
    
    cross = cross\
        .withColumn('0', F.col('0')/sum_0)\
        .withColumn('1', F.col('1')/sum_1)\
        .withColumn('woe', F.log(F.col('0')/F.col('1')))\
        .withColumn('iv', F.col('woe')*(F.col('0') - F.col('1')))\
        .withColumn('feature', F.lit(col_feature))\
        .fillna(0)

    iv = cross.selectExpr('sum(iv) as iv').collect()[0][0]  
        
    return cross.select('feature', 'feature_value', 'woe', 'iv'), iv


class WeightOfEvidenceComputer(
    Transformer,
    HasInputCol,
    HasOutputCol,
    HasInputCols,
    HasOutputCols,
    DefaultParamsReadable,
    DefaultParamsWritable
):
    """A transform to add the weight of evidence value for categorical variables.
    
    This is a class to add the columns with the Weight of Evidence for categorical features.
    See http://www.m-hikari.com/ams/ams-2014/ams-65-68-2014/zengAMS65-68-2014.pdf for a technical description.  
    """
    
    col_target = Param(
        parent=Params._dummy(),
        name='col_target',
        doc='Column name of the target. Must be a integer os values 0 or 1.',
        typeConverter=TypeConverters.toString
    )

    @keyword_only
    def __init__(
        self,
        inputCol=None,
        inputCols=None,
        col_target=None,
    ):
        super().__init__()
        self._setDefault(col_target=None)
        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    @keyword_only
    def setParams(
        self,
        inputCol=None,
        inputCols=None,
        col_target=None,
    ):
        kwargs = self._input_kwargs
        return self._set(**kwargs)

    def setTarget(self, new_target):
        return self.setParams(col_target=new_target)

    def setInputCol(self, new_inputCol):
        return self.setParams(inputCol=new_inputCol)

    def setInputCols(self, new_inputCols):
        return self.setParams(inputCols=new_inputCols)

    def getTarget(self):
        return self.getOrDefault(self.col_target)
    
    def checkParams(self):
        # Test #1: either inputCol or inputCols can be set (but not both).
        if self.isSet("inputCol") and (self.isSet("inputCols")):
            raise ValueError("Only one of `inputCol` and `inputCols`" "must be set.")

        # Test #2: at least one of inputCol or inputCols must be set.
        if not (self.isSet("inputCol") or self.isSet("inputCols")):
            raise ValueError("One of `inputCol` or `inputCols` must be set.")
    
    @typechecked
    def add_woe(self, dfs: DataFrame, feature: str, col_target: str) -> DataFrame:
        """Function that add teh WOE to the dataset passed to the tranform method of the class.

        Args:
            dfs (DataFrame): A spark DataFrame with feature and target columns.
            feature (str): Column name of a categorical feature, must be of type string.
            col_target (str): Column name of a target variable, must be of type integer with values 0 and 1.

        Returns:
            DataFrame: The dfs argument with a column f'{feature}_woe'.
        """

        dfs_woe_feature, iv = compute_woe_iv(dfs=dfs,col_feature=feature,col_target=col_target)
        dfs_woe_feature = dfs_woe_feature\
            .withColumnRenamed('feature_value', feature)\
            .withColumnRenamed('woe', f'{feature}_woe')\
            .select(feature, f'{feature}_woe')


        cols = dfs.columns + [f'{feature}_woe']
        dfs = dfs.join(dfs_woe_feature, on=feature, how='left').select(cols)
        return dfs

    def _transform(self, dataset):
        self.checkParams()

        # If `inputCol`, we wrap into a single-item list
        input_columns = (
            [self.getInputCol()]
            if self.isSet("inputCol")
            else self.getInputCols()
        )

        for feat in input_columns:
            dataset = self.add_woe(dfs=dataset, feature=feat, col_target=self.getTarget())
            
        return dataset

@typechecked
def feature_selection_with_iv(
    dfs: DataFrame,
    col_target: str,
    num_features: Union[List[str], None],
    cat_features: Union[List[str], None],
    floor_iv: float = 0.3,
    bucket_fraction: float = 0.1,
    categorical_as_woe: Union[bool, None] = False
) -> dict:
    """Function that executes a feature selection based on the information value methodology.

    This function computes the information value for the features passed and based on the floor_iv select the features
    that should ne used in the modeling part. It return analytical dataframes (with the Weight of Evidence and Information
    Value). 

    But the main advantage is that it also returns a list os stages to be passed to a pyspark.ml.Pipeline that will encode 
    and assemble the select variables.


    See http://www.m-hikari.com/ams/ams-2014/ams-65-68-2014/zengAMS65-68-2014.pdf for a technical description
    and https://www.listendata.com/2015/03/weight-of-evidence-woe-and-information.html for a introduction on the subject.

    Args:
        dfs (DataFrame): A spark DataFrame.
        col_target (str): Column name of a binary target column. Must be of integer type and have only value 0 and 1.
        num_features (Union[List[str], None]): List of columns names of numeric features. This column will be transformed with QuantileDiscretizer.
        cat_features (Union[List[str], None]): List of columns names of categorical features;
        floor_iv (float, optional): Threshold for a feature to be selected, greater or equal this value. Defaults to 0.3.
        bucket_fraction (float, optional): Fraction of the dataset to be used to create the buckets Must be between 0.05 and 0.5. Defaults to 0.1. 
        categorical_as_woe (bool, optional): If True categorical variables will as Weight of Evidence. If False one hot encoding. Defaults to False.

    Raises:
        TypeError: if (num_features is None) and (cat_features is None)
        ValueError: if (bucket_fraction < 0.05) or (bucket_fraction > 0.5)

    Returns:
        dict: A dict with the following structure
            - dfs_woe: Spark DataFrame with WOE and IV for each feature value. Given by the cols: feature, feature_value, woe, iv.
            - dfs_iv: Spark DataFrame with IV for each feature. Given by the cols: feature, iv.
            - stages_features_vector: List with spark transformers that computes the features vector based on the floor_iv and categorical_as_woe params.
    """
    # 1) Checks conditions
    if (num_features is None) and (cat_features is None):
        raise TypeError('num_features or cat_features must be a List[str]. Both are none.')
    
    if (bucket_fraction < 0.05) or (bucket_fraction > 0.5):
        raise ValueError('Param floor_bucket_fraction must be between 0.05 and 0.5')

    if (type(categorical_as_woe) is bool) and (cat_features is None):
        raise Warning('cat_features is None and categorical_as_woe is bool. categorical_as_woe param will have no effect.')

    # 2) If dfs has numerical features we need to bucket it!
    if num_features is not None:
        count_dfs = dfs.count()
        nBuckets = int(round(count_dfs/(count_dfs*bucket_fraction), 0))

        bucket_num_features = [i + '_bucket' for i in num_features]
        qt = QuantileDiscretizer(inputCols=num_features, outputCols=bucket_num_features, numBuckets=nBuckets)
        dfs = qt.fit(dfs).transform(dfs)
    
    # 3) Computing the WOE and IV
    ## 3a) Getting the column names to compute the WOE and IV
    if (num_features is not None) and (cat_features is not None):
        feats = bucket_num_features + cat_features
    elif num_features is None:
        feats = cat_features
    else:
        feats = bucket_num_features

    ## 3b) Computing the WOE and IV
    sc = SparkSession.builder.appName('spark_session_getter').getOrCreate()
    schema_woe = T.StructType([
        T.StructField("feature", T.StringType(), False),
        T.StructField("feature_value", T.StringType(), True),
        T.StructField("woe", T.FloatType(), True),
        T.StructField("iv", T.FloatType(), True)
    ])
    dfs_woe = sc.createDataFrame([], schema_woe)


    schema_iv = T.StructType([
        T.StructField('feature', T.StringType(), False),
        T.StructField('iv', T.FloatType(), False)
    ])
    dfs_iv = sc.createDataFrame([], schema_iv)

    for f in feats:
        df_woe_feature, iv = compute_woe_iv(dfs=dfs,col_feature=f,col_target=col_target)
    
        dfs_woe = dfs_woe.union(df_woe_feature)
        dfs_iv = dfs_iv.union(sc.createDataFrame(pd.DataFrame({'feature':[f],'iv':[iv]})))

    ## 3c) Selection features that have a IV greater then the floor_iv
    cols_to_keep = dfs_iv.filter(f'iv >= {floor_iv}').toPandas()['feature'].to_list()

    # 4) Creating the list with the transformation stages
    if (num_features is not None) and (cat_features is not None):
        cat_features_selected = list(filter(None, [None if s.endswith('_bucket') else s for s in cols_to_keep]))

        if categorical_as_woe:
            num_selected_features = [s[:-7] if s.endswith('_bucket') else s+'_woe' for s in cols_to_keep]
            # Creating WOEComputer Stage
            WOE = WeightOfEvidenceComputer(inputCols=cat_features_selected, col_target=col_target)
            # Creating list of Stages
            stages_features_vector = [WOE] + get_features_vector(num_features=num_selected_features)
        else:
            num_selected_features = list(filter(None, [s[:-7] if s.endswith('_bucket') else None for s in cols_to_keep]))
            stages_features_vector = get_features_vector(num_features=num_selected_features, cat_features=cat_features_selected)
            
    elif num_features is None:
        cat_features_selected = list(filter(None, [None if s.endswith('_bucket') else s for s in cols_to_keep]))
        if categorical_as_woe:
            num_selected_features = list(filter(None, [None if s.endswith('_bucket') else s+'_woe' for s in cols_to_keep]))
            
            # Creating WOEComputer Stage
            WOE = WeightOfEvidenceComputer(inputCols=cat_features_selected, col_target=col_target)
            # Creating list of Stages
            stages_features_vector = [WOE] + get_features_vector(num_features=num_selected_features)
        else:
            stages_features_vector = get_features_vector(cat_features=cat_features_selected)
    else:
        num_selected_features = list(filter(None, [s[:-7] if s.endswith('_bucket') else None for s in cols_to_keep]))
        stages_features_vector = get_features_vector(num_features=num_selected_features)

    # 5) Return object
    out_dict = {
        'dfs_woe': dfs_woe,
        'df_iv': dfs_iv.orderBy(F.col('iv').desc()).toPandas(),
        'stages_features_vector': stages_features_vector
    }

    return out_dict



"""Sub-package Dedicated to Data Wrangling tools.
"""

from pyspark_ds_toolbox.wrangling.reshape import pivot_long, with_start_week
from pyspark_ds_toolbox.wrangling.data_quality import count_percent_missing_rows_per_column


"""Difference in Difference toolbox.

For an introduction on the subject see: https://mixtape.scunning.com/difference-in-differences.html
"""

from typing import Union, List
from typeguard import typechecked

import pyspark
import pyspark.sql.functions as F
from pyspark.ml.regression import LinearRegression
from pyspark.ml import Pipeline

from pyspark_ds_toolbox.ml.data_prep import get_features_vector

@typechecked
def did_estimator(
    df: pyspark.sql.dataframe.DataFrame,
    id_col: str, 
    y: str, 
    flag_unit: str, 
    flag_time: str, 
    num_features: Union[None, List[str]] = None,
    cat_features: Union[None, List[str]] = None
) -> dict:
    """Difference in Difference Estimator.

    implementation based on https://matheusfacure.github.io/python-causality-handbook/14-Difference-in-Difference.html.
    
    Args:
        df (pyspark.sql.dataframe.DataFrame): SparkDF from which the causal effect will be estimated.
        id_col (str): Column name of an unique unit identifier.
        y (str): Column name of the outcome of interest.
        flag_unit (str): Column name of a flag indicating wheter the unit was treated or not. MUST be 1 or 0.
        flag_time (str): Column name of a flag indicating whether the time is before or after the treatment. MUST BE 1 or 0.
        num_features ([type], optional): List of numerics features to be used. Defaults to Union[None, List[str]].
        cat_features ([type], optional): List of categorical features to be used. Defaults to None.

    Raises:
        ValueError: If id_col, y, flag_unit or flag_time is not in df.columns.

    Returns:
        [dict]: A dictionary with the following keys and values
            - 'impacto_medio': list(linear_model.coefficients)[0],
            - 'n_ids_impactados': df_model.select(id_col).distinct().count(),
            - 'impacto': list(linear_model.coefficients)[0]*df_model.select(id_col).distinct().count(),
            - 'pValueInteraction': linear_model.summary.pValues[0],
            - 'r2': linear_model.summary.r2,
            - 'r2adj': linear_model.summary.r2adj,
            - 'df_with_features': df_model,
            - 'linear_model': linear_model
    """

    for n in [id_col, y, flag_unit, flag_time]:
        if n not in df.columns:
            raise ValueError('id_col, y, flag_unit, flag_time must be columns of df.')

    df_model = df.withColumn('interaction', F.col(flag_unit)*F.col(flag_time))

    if (num_features is None) and (cat_features is None):
        num_features=['interaction', flag_unit, flag_time]
        cat_features=None
    elif (num_features is not None) and (cat_features is None):
        num_features=['interaction', flag_unit, flag_time] + num_features
        cat_features=None
    elif (num_features is None) and (cat_features is not None):
        num_features=['interaction', flag_unit, flag_time]
    else:
        pass

    stages = get_features_vector(
        num_features=num_features,
        cat_features=cat_features
    )

    pipeline = Pipeline(stages=stages+[LinearRegression(
                                                        featuresCol = 'features', 
                                                        labelCol='deposits', 
                                                        fitIntercept=True
                                                        )
                                        ]
                        )
    fitted_pipeline = pipeline.fit(df_model)
    linear_model = fitted_pipeline.stages[-1]
    
    out_dict = {
        'impacto_medio': list(linear_model.coefficients)[0],
        'n_ids_impactados': df_model.select(id_col).distinct().count(),
        'impacto': list(linear_model.coefficients)[0]*df_model.select(id_col).distinct().count(),
        'pValueInteraction': linear_model.summary.pValues[0],
        'r2': linear_model.summary.r2,
        'r2adj': linear_model.summary.r2adj,
        'df_with_features': df_model,
        'linear_model': linear_model
    }
    return out_dict



"""Module dedicated to features spark vector tools.
"""

from typing import List, Union
from typeguard import typechecked

import pyspark 
import pyspark.ml.feature as FF

@typechecked
def get_features_vector(
    num_features: Union[List[str], None] = None,
    cat_features: Union[List[str], None] = None,
    output_col = 'features'
) -> List:
    """Assembles a features vector to be used with ML algorithms.

    Args:
        num_features (List[str]): List of columns names of numeric features;
        cat_features (List[str]): List of column names of categorical features (StringIndexer);
        output_col (str): name of the output column;

    Raises:
        TypeError: If num_features AND cat_features are os type None.

    Returns:
        [List]: pyspark indexers, encoders and assemblers like a list;
    """
    if (num_features is None) and (cat_features is None):
        raise TypeError('num_features or cat_features must be a List[str]. Both are none.')
    
    if cat_features is not None:
        indexers = [FF.StringIndexer(inputCol = c, outputCol="{0}_indexed".format(c)) for c in cat_features]
        encoders = [FF.OneHotEncoder(inputCol = indexer.getOutputCol(), outputCol = "{0}_encoded".format(indexer.getOutputCol())) for indexer in indexers]
        assemblerCat = FF.VectorAssembler(inputCols = [encoder.getOutputCol() for encoder in encoders], outputCol = "cat")
        stagesCat = indexers + encoders + [assemblerCat]
    
    if num_features is not None:
        assemblerNum = FF.VectorAssembler(inputCols = num_features, outputCol = "num")
        stagesNum = [assemblerNum]

    if (num_features is not None) and (cat_features is not None):
        columns = ['num', 'cat']
        stages = stagesCat + stagesNum
    elif num_features is None:
        columns = ['cat']
        stages = stagesCat
    else:
        columns = ['num']
        stages = stagesNum
    
    stagesList = stages + [FF.VectorAssembler(inputCols = columns, outputCol = output_col)]

    return stagesList



__doc__  = """pyspark_ds_toobox

A Pyspark companion for data science tasks.
"""

__version__ = '0.4.2'

from pyspark_ds_toolbox import wrangling
from pyspark_ds_toolbox import ml
from pyspark_ds_toolbox import causal_inference
from pyspark_ds_toolbox import stats


"""Module Dedicated to Association Metrics

The class implemented in this module is based on the book:
Morettin, P.A. and Bussab, W.O., 2017. Estatística básica. Saraiva Educação SA.
"""

import seaborn as sns
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

import pyspark
import pyspark.pandas as ps
from typeguard import typechecked
from typing import List, Union

class Association():
    """This association class implements different types of association metrics, for both categorical and numerical variables.
    
    The class implemented in this module is based on the book:
    Morettin, P.A. and Bussab, W.O., 2017. Estatística básica. Saraiva Educação SA.

    The current implementation is built on top of Koalas, but in the future this will be change to pure pyspark.
    """
    @typechecked
    def C(
        self, 
        df: pyspark.pandas.frame.DataFrame, 
        columns: List[str], 
        dense: bool=True
    ) -> float:
        """Computes the Contingency Coefficient.
        A non-normalized association metric between two categorical variables.

        Args:
            df (pyspark.pandas.frame.DataFrame): A PandasOnSParkDF (sparkDF.to_pandas_on_spark()) contaning the data in dense form or in grouped form.
            columns (List[str]): List of Strings containing the name of the categorical columns which will be used in the coefficient computation.
            dense (bool, optional): If false it is expected that you have executed this df.groupby(columns).size().unstack(level=0).fillna(0) in the df.
                Defaults to True.

        Returns:
            float: The Contingency Coefficient value.
        """
        if (dense):
            num_ocorrencias = df.groupby(columns).size().unstack(level=0).fillna(0)
        else:
            num_ocorrencias = df

        n = num_ocorrencias.sum().sum()
        prob_marginal = num_ocorrencias.sum(axis=0)/num_ocorrencias.sum().sum()
        total_marginal = num_ocorrencias.sum(axis=1)
        valores_esperados = num_ocorrencias.copy()
        for i in ps.DataFrame(total_marginal).iterrows():
            for j in ps.DataFrame(prob_marginal).iterrows():
                valores_esperados.loc[i[0], j[0]] = total_marginal[i[0]]*prob_marginal[j[0]]

        qui_quadrado = ((num_ocorrencias - valores_esperados).pow(2)/valores_esperados).sum().sum()
        C = np.sqrt(qui_quadrado/(qui_quadrado + n))
        return C
    
    @typechecked
    def T(
        self, 
        df: pyspark.pandas.frame.DataFrame, 
        columns: List[str], 
        dense: bool=True
    ) -> float:
        """Computes the T Metrics.
        A normalized metric of assiciation between two categorical variables.

        Args:
            df (pyspark.pandas.frame.DataFrame): A PandasOnSParkDF (sparkDF.to_pandas_on_spark()) contaning the data in dense form or in grouped form.
            columns (List[str]): List of Strings containing the name of the categorical columns which will be used in the coefficient computation.
            dense (bool, optional): If false it is expected that you have executed this df.groupby(columns).size().unstack(level=0).fillna(0) in the df.

        Returns:
            float: The T Metric.
        """
        if (dense):
            num_ocorrencias = df.groupby(columns).size().unstack(level=0).fillna(0)
        else:
            num_ocorrencias = df
        
        r, s = num_ocorrencias.shape
        n = num_ocorrencias.sum().sum()
        prob_marginal = num_ocorrencias.sum(axis=0)/num_ocorrencias.sum().sum()
        total_marginal = num_ocorrencias.sum(axis=1)
        valores_esperados = num_ocorrencias.copy()
        
        for i in ps.DataFrame(total_marginal).iterrows():
            for j in ps.DataFrame(prob_marginal).iterrows():
                valores_esperados.loc[i[0], j[0]] = total_marginal[i[0]]*prob_marginal[j[0]]

        qui_quadrado = ((num_ocorrencias - valores_esperados).pow(2)/valores_esperados).sum().sum()
        T = np.sqrt((qui_quadrado/n)/(np.sqrt((r-1)*(s-1))))
        return T
    
    @typechecked
    def corr(
        self, 
        df: pyspark.pandas.frame.DataFrame, 
        columns: List[str]
    ) -> float:
        """Computes the Correlation Coefficient.

        Standard Correlation Coefficient.

        Args:
            df (databricks.koalas.frame.DataFrame): A PandasOnSParkDF (sparkDF.to_pandas_on_spark()).
            columns (List[str]): List of numeric column names from which the correlation will computed.

        Returns:
            float: Correlation Coefficient.
        """
        X = df[columns[0]]
        Y = df[columns[1]]
      
        n = X.shape[0]
        X_mean = X.mean()
        Y_mean = Y.mean()
        
        num = (X*Y).sum() - n*X_mean*Y_mean
        den = np.sqrt(((X*X).sum() - n*X_mean*X_mean) * ((Y*Y).sum() - n*Y_mean*Y_mean))
        return num/den
    
    @typechecked
    def R2(
        self, 
        df: pyspark.pandas.frame.DataFrame, 
        categorical: str, 
        numerical: str
    ) -> float:
        """Computes the R2 Metric for one numeric column and one categorical column.

        Metric of association between a numeric and a categorical variables.

        Args:
            df (pyspark.pandas.frame.DataFrame): A PandasOnSParkDF (sparkDF.to_pandas_on_spark()).
            categorical (str): Column name of the categorical column.
            numerical (str): column name of the numerical column.

        Returns:
            float: The R2 Metric.
        """
        # X = df[categorical]
        Y = df[numerical]

        data = df.loc[:, [categorical, numerical]]
        Y_var = Y.var()
        Yi_var = data.groupby(categorical).agg({numerical: ['variance', 'count']})

        Y_var__ = (Yi_var[numerical]['variance']*Yi_var[numerical]['count']).sum()/Yi_var[numerical]['count'].sum()
        R2 = 1 - Y_var__/Y_var
        return R2  
    
    @typechecked
    def association_matrix(
        self, 
        df: pyspark.pandas.frame.DataFrame, 
        categorical_features: Union[List[str], None] = None, 
        numerical_features: Union[List[str], None] = None,
        plot_matrix: bool = True,
        return_matrix: bool = False
    ) -> Union[None, pd.core.frame.DataFrame]:
        """Computes from a df, a list of categorical and a list of numerical variables a normalized association matrix.

        Args:
            df (pyspark.pandas.frame.DataFrame): A PandasOnSParkDF (sparkDF.to_pandas_on_spark()). 
            categorical_features (List[str]): List of column names of the categorical features.
            numerical_features (List[str]): List of Column names of the numerical features.
            plot_matrix (bool, optional): If set False it will not plot the matrix. Defaults to True.
            return_matrix (bool, optional): If set to True it will return the correlation matrix as a pandasDF. Defaults to False.

        raises:
            ValueError: if (categorical_features is None) and (numerical_features is None) is True

        Returns:
            Union[None, pd.core.frame.DataFrame]: Either None, if return_matrix is False, or a PandasDF with the correlation coefficients.
        """

        if (categorical_features is None) and (numerical_features is None):
            raise ValueError('Both categorical_features and numerical_features are of type None. At least one must be List[str].')
        
        if categorical_features is None:
            features = numerical_features
        elif numerical_features is None:
            features = categorical_features
        else:
            features = categorical_features + numerical_features
        
        aMatrix = ps.DataFrame(columns=features, index=features, dtype=np.float64)
        
        for i in aMatrix.iterrows():
            for j in aMatrix.iteritems():
                if ((i[0] in numerical_features) and j[0] in (numerical_features)):
                    aMatrix.loc[i[0], j[0]] = self.corr(df, [i[0], j[0]])
                elif((i[0] in categorical_features) and j[0] in (categorical_features)):
                    aMatrix.loc[i[0], j[0]] = self.T(df, [i[0], j[0]])
                elif((i[0] in categorical_features) and j[0] in (numerical_features)):
                    aMatrix.loc[i[0], j[0]] = self.R2(df, i[0], j[0])
                elif((i[0] in numerical_features) and j[0] in (categorical_features)):
                    aMatrix.loc[i[0], j[0]] = self.R2(df, j[0], i[0])
        
        if plot_matrix:
            fig, ax = plt.subplots(figsize=(15, 7))
            sns.heatmap(aMatrix.to_pandas().round(2), xticklabels=aMatrix.columns, yticklabels=aMatrix.columns, ax=ax, annot=True)
            fig.show()

        if (return_matrix):
            return aMatrix.to_pandas()



"""Evaluation toolbox.

Module dedicated to functionalities related to classification evaluation.
"""

import sys
from typing import List
from typeguard import typechecked
import pandas as pd

import pyspark 
from pyspark.sql import functions as F, types as T, Window
from pyspark.ml.evaluation import BinaryClassificationEvaluator
from pyspark.ml.linalg import Vectors, VectorUDT


get_p1 = F.udf(lambda value: value[1].item(), T.FloatType())

@typechecked
def binary_classificator_evaluator(
    dfs_prediction: pyspark.sql.dataframe.DataFrame,
    col_target: str,
    col_prediction: str,
    print_metrics: bool = False
) -> dict:
    """Computes the Matrics of a Binary Classifier from a Prediction Output table from spark.

    Args:
        dfs_prediction (pyspark.sql.dataframe.DataFrame): Output Prediction table from spark binarry classifier.
        col_target (str): Column name with the target (ground truth)
        print_metrics (bool, optional): Wether to print or not the metrics in the console. Defaults to False.

    Raises:
        Exception: Any error that is encontered.

    Returns:
        [Dict]: Dict with the following metrics
            - confusion_matrix
            - accuracy
            - precision
            - recall
            - f1 score
            - aucaoc
    """
    try:
        # Confusion Matrix
        confusion_matrix = dfs_prediction.groupBy(col_target, col_prediction).count() 
        TN = dfs_prediction.filter(f'{col_prediction} = 0 AND {col_target} = 0').count()
        TP = dfs_prediction.filter(f'{col_prediction} = 1 AND {col_target} = 1').count()
        FN = dfs_prediction.filter(f'{col_prediction} = 0 AND {col_target} = 1').count()
        FP = dfs_prediction.filter(f'{col_prediction} = 1 AND {col_target} = 0').count()

        # Computing Metrics from Confusion Matrix
        accuracy = (TN + TP) / (TN + TP + FN + FP)
        precision = TP/(TP+FP) if (TP+FP) > 0 else 0.0
        recall = TP/(TP+FN)
        f1 = 2*(precision*recall/(precision+recall)) if (precision+recall) > 0 else 0.0 

        evaluator = BinaryClassificationEvaluator(labelCol=col_target, rawPredictionCol=col_prediction, metricName='areaUnderROC')
        aucroc = evaluator.evaluate(dfs_prediction)
        evaluator = BinaryClassificationEvaluator(labelCol=col_target, rawPredictionCol=col_prediction, metricName='areaUnderPR')
        aucpr = evaluator.evaluate(dfs_prediction)

        # Results Dict
        out_dict = {
            'confusion_matrix': confusion_matrix.toPandas(),
            'accuracy': accuracy,
            'f1': f1,
            'precision': precision,
            'recall': recall,
            'aucroc': aucroc,
            'aucpr': aucpr
        }
        
        if print_metrics == True:
            print(f'accuracy:  {round(out_dict["accuracy"], 4)}')
            print(f'precision: {round(out_dict["precision"], 4)}')
            print(f'recall:    {round(out_dict["recall"], 4)}')
            print(f'f1:        {round(out_dict["f1"], 4)}')
            print(f'auroc:     {round(out_dict["auroc"], 4)}')
        
        return out_dict
    except Exception as e:
        raise Exception(e)

@typechecked
def binary_classifier_decile_analysis(
    dfs: pyspark.sql.dataframe.DataFrame,
    col_id: str,
    col_target: str,
    col_probability: str
) -> pyspark.sql.dataframe.DataFrame:
    """Computes a Precision, Recall and KS decile analysis from a probability prediction model.
    col_target column MUST have values of only 0 and 1.

    Args:
        dfs (pyspark.sql.dataframe.DataFrame): SparkDF with probabilities predictions.
        col_id (str): Column name with id value, to count the values for each decile.
        col_target (str): Column name with the ground truth. Must be from a binary classifier with values 1 and 0.
        col_probability (str): Column name with the probability estimated from the model.

    Raises:
        ValueError: If unique values from col_target column are not 0 and 1.

    Returns:
        pyspark.sql.dataframe.DataFrame: SparkDF with the columns:
            - percentile, min_prob, max_prob, count_id, events, non_events, cum_events,
            cum_non_events, precision_at_percentile, recall_at_percentile, event_rate,
            nonevent_rate, cum_eventrate, cum_noneventrate, ks.
    """
    unique_rows = dfs.select(col_target).distinct().collect()
    unique_rows_values = [v[col_target] for v in unique_rows]
    unique_rows_values.sort()
    
    if unique_rows_values != [0, 1]:
        raise ValueError(f'Unique values from {col_target} column are not [0, 1].')

    # 1 Adding non events column to dataset
    dfs = dfs.withColumn('target0',1 - dfs[col_target])
    
    # 2 Adding the decile to the dataset
    dfs = dfs.withColumn("prob_qcut", F.ntile(10).over(Window.partitionBy().orderBy(dfs[col_probability])))
    
    # 3 Aggragating to compute the counts and min/max probabilities
    decile_table = dfs\
            .groupBy('prob_qcut')\
            .agg(
                F.count(col_id),
                F.avg(col_probability),
                F.min(col_probability),
                F.max(col_probability),
                F.sum(col_target),
                F.sum('target0')
            )\
            .orderBy(F.col(f'min({col_probability})').desc())\
            .withColumnRenamed(f'count({col_id})', 'count_id')\
            .withColumnRenamed(f'avg({col_probability})', 'avg_prob')\
            .withColumnRenamed(f'min({col_probability})', 'min_prob')\
            .withColumnRenamed(f'max({col_probability})', 'max_prob')\
            .withColumnRenamed(f'sum({col_target})', 'events')\
            .withColumnRenamed('sum(target0)','non_events')
    
    # 4 Computing the events and non-events rates
    count_event = dfs.filter(f'{col_target} == 1').count()
    count_nonevent = dfs.filter(f'{col_target} == 0').count()
    decile_table = decile_table.withColumn('event_rate', decile_table.events/count_event)
    decile_table = decile_table.withColumn('nonevent_rate', decile_table.non_events/count_nonevent)
    
    # 5 Cumulating the events/non-evenys values and rates
    win = Window.partitionBy().orderBy().rowsBetween(-sys.maxsize, 0)
    decile_table = decile_table.withColumn('cum_events', F.sum(decile_table.events).over(win))
    decile_table = decile_table.withColumn('cum_non_events', F.sum(decile_table.non_events).over(win))
    decile_table = decile_table.withColumn('cum_eventrate', F.sum(decile_table.event_rate).over(win)) # recall
    decile_table = decile_table.withColumn('cum_noneventrate', F.sum(decile_table.nonevent_rate).over(win))
    
    # 6 Compute the precision at percentiles
    decile_table = decile_table.withColumn('precision_at_percentile', F.col('cum_events')/(F.col('cum_events')+F.col('cum_non_events')))

    # 7 Computing the KS Metric
    decile_table = decile_table.withColumn('ks', F.expr('cum_eventrate - cum_noneventrate'))
    
    # 8 Adding decile number
    decile_table = decile_table.withColumn('percentile', F.row_number().over(Window.orderBy(F.lit(1))))

    # 9 Selecting columns of interest
    decile_table = decile_table.select(
        'percentile', 'min_prob', 'max_prob', 'avg_prob', 'count_id', 'non_events', 'events', 
        'cum_non_events', 'cum_events', 'nonevent_rate', 'event_rate', 'cum_noneventrate',
        'cum_eventrate',  'precision_at_percentile', 'ks'
    )
    return decile_table



"""Module dedicated to functionalities related to class weighting tools.
"""

from typeguard import typechecked

import pyspark 
from pyspark.sql import functions as F


@typechecked
def binary_classifier_weights(dfs: pyspark.sql.dataframe.DataFrame, col_target: str) -> pyspark.sql.dataframe.DataFrame:
    """Adds a class weight columns to a binary classification response column.

    Args:
        dfs (pyspark.sql.dataframe.DataFrame): Training dataset with the col_target column.
        col_target (str): Column name of the column that contains the response variable for the model.
            It should contain only values of 0 and 1.


    Raises:
        ValueError: If unique values from col_target column are not 0 and 1.

    Returns:
        pyspark.sql.dataframe.DataFrame: The dfs object with a weight_{col_target} column.
    """
    # 1) Check if col_target unique values are [0, 1] 
    unique_rows = dfs.select(col_target).distinct().collect()
    unique_rows_values = [v[col_target] for v in unique_rows]
    unique_rows_values.sort()
    if unique_rows_values != [0, 1]:
        raise ValueError(f'Unique values from {col_target} column are not [0, 1].')
    
    # 2) Computes the class weights
    count_parc = dfs.filter(f'{col_target} = 1').count()
    count_total = dfs.count()
    c = 2
    weight_parc = count_total / (c * count_parc)
    weight_no_parc = count_total / (c * (count_total - count_parc)) 

    # 3) Adds the weight column to dfs
    df_with_weight_columns = dfs.withColumn(f"weight_{col_target}", F.when(F.col(col_target) == 1, weight_parc).otherwise(weight_no_parc))

    return df_with_weight_columns



__doc__  = """Machine Learning toolbox.

Subpackage dedicated to Machine Learning helpers.
"""

from pyspark_ds_toolbox.ml import data_prep
from pyspark_ds_toolbox.ml.classification import eval, baseline_classifiers
from pyspark_ds_toolbox.ml.feature_importance import shap_values


"""Feature Selection tools.

Subpackage dedicated to Feature Selection tools.
"""

from pyspark_ds_toolbox.ml.feature_selection.information_value import feature_selection_with_iv,\
    compute_woe_iv, WeightOfEvidenceComputer